# status used for logging and jobs
STATUS_CREATED = "created"
STATUS_RUNNING = "running"
STATUS_CANCELED = "canceled"
STATUS_COMPLETED = "completed"
STATUS_FAILED = "failed"
STATUS_ALL = (STATUS_CREATED, STATUS_RUNNING, STATUS_CANCELED, STATUS_COMPLETED, STATUS_FAILED)
